from pysimmmulator.param_handlers import (
    basic_parameters,
    baseline_parameters,
    ad_spend_parameters,
    media_parameters,
    cvr_parameters,
    adstock_parameters,
    output_parameters
)

import numpy as np
import pandas as pd

import logging
import logging.config

logging.config.fileConfig("./logging.conf")
logger = logging.getLogger("pysimmmulator")


class simulate:
    """Takes input of basic params and provies either piece meal or single shot 
    creation of MMM data using a config file,"""
    def __init__(self, basic_params: basic_parameters):
        self.basic_params = basic_params

    def simulate_baseline(
        self,
        base_p,
        trend_p: int,
        temp_var: int,
        temp_coef_mean: int,
        temp_coef_sd: int,
        error_std: int,
    ) -> None:
        self.baseline_params = baseline_parameters(
            basic_params=self.basic_params,
            base_p=base_p,
            trend_p=trend_p,
            temp_var=temp_var,
            temp_coef_mean=temp_coef_mean,
            temp_coef_sd=temp_coef_sd,
            error_std=error_std,
        )

        # Number of days to generate data for
        days = np.arange(0, self.basic_params.years * 365)
        # Base sales of base_p units
        base = np.zeros(shape=self.basic_params.years * 365) + self.baseline_params.base_p
        # Trend of trend_p extra units per day
        trend_cal = (self.baseline_params.trend_p / (self.basic_params.years * 365)) * self.baseline_params.base_p
        trend = trend_cal * days
        # Temperature generated by a sin function and we can manipulate how much the sin function goes up or down with temp_var
        temp = self.baseline_params.temp_var * np.sin(days * 3.14 / 182.5)
        # coefficient of temperature's effect on sales will be a random variable with normal distribution
        seasonality = (
            np.random.normal(loc=self.baseline_params.temp_coef_mean, scale=self.baseline_params.temp_coef_sd, size=1)
            * temp
        )
        # add some noise to the trend
        error = np.random.normal(
            loc=0, scale=self.baseline_params.error_std, size=self.basic_params.years * 365
        )
        # Generate series for baseline sales
        baseline_sales = base + trend + seasonality + error
        # if error term makes baseline_sales negative, make it 0
        baseline_sales = np.where(baseline_sales < 0, 0, baseline_sales)

        self.baseline_sales_df = pd.DataFrame(
            {
                "days": days,
                "baseline_sales": baseline_sales,
                "base": base,
                "trend": trend,
                "temp": temp,
                "seasonality": seasonality,
                "error": error,
            }
        )

    def simulate_ad_spend(
        self,
        campaign_spend_mean: int,
        campaign_spend_std: int,
        max_min_proportion_on_each_channel: dict,
    ) -> None:
        ad_spend_params = ad_spend_parameters(
            campaign_spend_mean=campaign_spend_mean,
            campaign_spend_std=campaign_spend_std,
            max_min_proportion_on_each_channel=max_min_proportion_on_each_channel,
        )

        campaign_count = int(
            self.basic_params.years * 365 / self.basic_params.frequency_of_campaigns
        )

        # specify amount spent on each campaign according to a normal distribution
        campaign_spends = np.round(
            np.random.normal(
                loc=ad_spend_params.campaign_spend_mean,
                scale=ad_spend_params.campaign_spend_std,
                size=campaign_count,
            ), 2
        )
        # if campaign spend number is negative, automatically make it 0
        campaign_spends[campaign_spends < 0] = 0
        campaign_channel_spend_proportions = {}
        for (
            channel,
            proportions,
        ) in ad_spend_params.max_min_proportion_on_each_channel.items():
            campaign_channel_spend_proportions[channel] = np.random.uniform(
                low=proportions["min"],
                high=proportions["max"],
                size=campaign_count,
            )

        spend_df = pd.DataFrame(
            {
                "campaign_id": np.arange(campaign_count),
                "total_campaign_spend": campaign_spends,
            }
        )

        for channel in max_min_proportion_on_each_channel.keys():
            spend_df[channel] = np.round(
                campaign_spends * campaign_channel_spend_proportions[channel], 2
            )

        self.spend_df = spend_df.melt(
            id_vars=["campaign_id", "total_campaign_spend"],
            value_vars=self.basic_params.all_channels,
            var_name="channel",
            value_name="spend_channel",
        )
        logger.info("You have completed running step 2: Simulating ad spend.")

    def simulate_media(
        self, true_cpm: dict, true_cpc: dict, noisy_cpm_cpc: dict
    ) -> None:
        media_params = media_parameters(true_cpm, true_cpc, noisy_cpm_cpc)
        media_params.check(basic_params=self.basic_params)
        
        for channel in media_params.noise_channels:
            channel_idx = self.spend_df[self.spend_df["channel"] == channel].index

            channel_noise = np.random.normal(size=len(channel_idx), **noisy_cpm_cpc[channel])

            channel_true_cpm_value = true_cpm[channel] if channel in true_cpm.keys() else np.nan
            channel_noisy_cpm_value = true_cpm[channel] + channel_noise if channel in true_cpm.keys() else np.nan
            self.spend_df.loc[channel_idx,'true_cpm'] = channel_true_cpm_value
            self.spend_df.loc[channel_idx,'noisy_cpm'] = channel_noisy_cpm_value

            channel_true_cpc_value = true_cpc[channel] if channel in true_cpc.keys() else np.nan
            channel_noisy_cpc_value = true_cpc[channel] + channel_noise if channel in true_cpc.keys() else np.nan
            self.spend_df.loc[channel_idx,'true_cpc'] = channel_true_cpc_value
            self.spend_df.loc[channel_idx,'noisy_cpc'] = channel_noisy_cpc_value

        self.spend_df['lifetime_impressions'] = np.round(self.spend_df['spend_channel'] / self.spend_df['noisy_cpm'] * 1000, 0)
        self.spend_df['lifetime_clicks'] = np.round(self.spend_df['spend_channel'] / self.spend_df['noisy_cpc'], 0)
        
        self.spend_df['daily_spend'] = np.round(self.spend_df['spend_channel'] / self.basic_params.frequency_of_campaigns, 2)
        self.spend_df['daily_impressions'] = np.round(self.spend_df['lifetime_impressions'] / self.basic_params.frequency_of_campaigns, 0)
        self.spend_df['daily_clicks'] = np.round(self.spend_df['lifetime_clicks'] / self.basic_params.frequency_of_campaigns, 0)

        logger.info("You have completed running step 3: Simulating media.")

    def simulate_cvr(self,  noisy_cvr: dict) -> None:
        cvr_params = cvr_parameters(noisy_cvr)
        cvr_params.check(basic_params=self.basic_params)
        
        for channel in cvr_params.noise_channels:
            channel_idx = self.spend_df[self.spend_df["channel"] == channel].index

            channel_noise = np.random.normal(size=len(channel_idx), **noisy_cvr[channel])
            self.spend_df.loc[channel_idx,'noisy_cvr'] = channel_noise + self.basic_params.true_cvr[channel]
        
        # Daily CVR == campaign CVR, no reason to duplicate
        logger.info("You have completed running step 4: Simulating CVR.")

    def _reformat_for_mmm(self) -> None:
        date_backbone = pd.date_range(start=self.basic_params.start_date, end=self.basic_params.end_date, freq='D')
        campaigns_in_period = date_backbone.shape[0] / self.basic_params.frequency_of_campaigns
        campaign_id_to_date_map = np.trunc(np.linspace(start=0, stop=campaigns_in_period-1, num=date_backbone.shape[0])).astype(int)
        self.mmm_df = pd.DataFrame({'date':date_backbone, 'id_map':campaign_id_to_date_map})
        self.mmm_df.set_index("id_map")
        
        agg_media_df = self.spend_df.groupby(['channel','campaign_id']).sum()[["daily_impressions","daily_clicks","daily_spend","noisy_cvr"]]
        agg_media_df = agg_media_df.unstack(level=0)
        joined_columns = []
        for (_metric, _channel) in agg_media_df.columns:
            # we'll just name everything channel_metric from here. No need for daily/lifetime
            col_name = f"{_channel}_{_metric.split('_')[1]}"
            joined_columns.append(col_name)
        agg_media_df.columns = joined_columns

        self.mmm_df = self.mmm_df.join(agg_media_df)
        del self.mmm_df["id_map"]

        logger.info("You have completed running step 5a: pivoting the data frame to an MMM format.")
    
    @staticmethod
    def _build_decay_vector(original_vector: pd.Series, decay_value: float) -> pd.Series:
        decayed_vector = [original_vector.values[0]]
        for i, orig_value in enumerate(original_vector.values[1:]):
            decayed_vector.append(orig_value + decay_value * decayed_vector[i])
        return pd.Series(decayed_vector)

    def _simulate_decay(self, true_lambda_decay: dict) -> None:
        for channel in true_lambda_decay.keys():
            metric = 'impressions' if channel in self.basic_params.channels_impressions else 'clicks'
            self.mmm_df[f"{channel}_{metric}_adstocked"] = self._build_decay_vector(original_vector=self.mmm_df[f"{channel}_{metric}"], decay_value=true_lambda_decay[channel])
        
        logger.info("You have completed running step 5b: applying adstock decay.")
        # Knew I could find a better way, even better now

    def _simulate_diminishing_returns(self, alpha_saturation: dict, gamma_saturation: dict
                                    #   , x_marginal: int = None
                                      ) -> None:
        for channel in alpha_saturation.keys():
            metric = 'impressions' if channel in self.basic_params.channels_impressions else 'clicks'
            target = self.mmm_df[f"{channel}_{metric}_adstocked"]
            gamma_trans = np.round(np.quantile(np.linspace(min(target), max(target), num=100), gamma_saturation[channel]), 4)
            x_scurve = target**alpha_saturation[channel] / (target**alpha_saturation[channel] + gamma_trans**alpha_saturation[channel])
            self.mmm_df[f"{channel}_{metric}_adstocked_decay_diminishing"] = x_scurve * target
        
        logger.info("You have completed running step 5c: apply diminishing marginal returns.")

    def simulate_decay_returns(self, true_lambda_decay: dict, alpha_saturation: dict, gamma_saturation: dict) -> None:
        adstock_params = adstock_parameters(true_lambda_decay, alpha_saturation, gamma_saturation)
        self._reformat_for_mmm()
        self._simulate_decay(adstock_params.true_lambda_decay)
        self._simulate_diminishing_returns(alpha_saturation = adstock_params.alpha_saturation,
                                           gamma_saturation = adstock_params.gamma_saturation)
        
        logger.info("You have completed running step 5: Simulating adstock.")

    def calculate_conversions(self):
        for channel in self.basic_params.all_channels:
            metric = 'impressions' if channel in self.basic_params.channels_impressions else 'clicks'
            self.mmm_df[f'{channel}_conversions'] = self.mmm_df[f'{channel}_{metric}_adstocked_decay_diminishing'] * self.mmm_df[f'{channel}_cvr']
        
        logger.info("You have completed running step 6: Calculating the number of conversions.")

    def consolidate_dataframe(self):
        all_cols = self.mmm_df.columns.tolist()
        metric_cols = []
        [metric_cols.append(f"{channel}_impressions") for channel in self.basic_params.channels_impressions]
        [metric_cols.append(f"{channel}_clicks") for channel in self.basic_params.channels_clicks]
        spend_cols = []
        [spend_cols.append(f"{channel}_spend") for channel in self.basic_params.all_channels]
        conv_cols = []
        [conv_cols.append(f"{channel}_conversions") for channel in self.basic_params.all_channels]
        self.mmm_df = self.mmm_df[metric_cols + spend_cols + conv_cols]
        self.mmm_df["total_conversions_from_ads"] = self.mmm_df[conv_cols].sum(axis=1)
        self.mmm_df["total_revenue_from_ads"] = self.mmm_df["total_conversions_from_ads"] * self.basic_params.revenue_per_conv
        self.mmm_df["baseline_revenue"] = round(self.baseline_sales_df["baseline_sales"]) * self.basic_params.revenue_per_conv
        self.mmm_df["total_revenue"] = self.mmm_df[["total_revenue_from_ads","baseline_revenue"]].sum(axis=1)
        
        logger.info("You have completed running step 7: Expanding to maximum data frame.")
    
    def calculate_channel_roi(self):
        self.channel_roi = {}
        for channel in self.basic_params.all_channels:
            total_cpa = self.mmm_df[f"{channel}_spend"].sum() / self.mmm_df[f"{channel}_conversions"].sum()
            total_roi = (self.basic_params.revenue_per_conv - total_cpa) / total_cpa
            self.channel_roi[channel] = total_roi
    
    def finalize_output(self, aggregation_level: str) -> None:
        output_params = output_parameters(aggregation_level)
        metric_cols = []
        [metric_cols.append(f"{channel}_impressions") for channel in self.basic_params.channels_impressions]
        [metric_cols.append(f"{channel}_clicks") for channel in self.basic_params.channels_clicks]
        spend_cols = []
        [spend_cols.append(f"{channel}_spend") for channel in self.basic_params.all_channels]
        
        if aggregation_level == "daily":
            self.final_df = self.mmm_df[metric_cols + spend_cols + ["total_revenue"]]
        else:
            self.mmm_df['week_start'] = self.mmm_df['date'] - pd.to_timedelta(self.mmm_df['date'].apply(lambda x: x.weekday()), unit='d')
            self.final_df = self.mmm_df.groupby(['week_start']).sum()[metric_cols + spend_cols + ["total_revenue"]]
        
        logger.info("You have completed running step 9: Finalization of output dataframe at the {aggregation_level} level")


    def run_with_config(self):
        import pysimmmulator.load_parameters as load_params
        self.simulate_baseline(**load_params.cfg['baseline_params'])
        self.simulate_ad_spend(**load_params.cfg["ad_spend_params"])
        self.simulate_media(**load_params.cfg["media_params"])
        self.simulate_cvr(**load_params.cfg['cvr_params'])
        self.simulate_decay_returns(**load_params.cfg["adstock_params"])
        self.calculate_conversions()
        self.consolidate_dataframe()



import yaml
from pysimmmulator.param_handlers import (
    basic_parameters,
    baseline_parameters,
    ad_spend_parameters,
    media_parameters,
    cvr_parameters,
    adstock_parameters,
    output_parameters
)


def load_config(config_path: str = "config.yaml") -> dict:
    with open(config_path) as f:
        cfg = yaml.load(f, Loader=yaml.FullLoader)
    return cfg


def build_params(config_dict: dict) -> None:
    my_basic_params = basic_parameters(**config_dict["basic_params"])
    my_baseline_params = baseline_parameters(
        basic_params=my_basic_params, **config_dict["baseline_params"]
    )


cfg = load_config(config_path="config.yaml")
my_basic_params = basic_parameters(**cfg["basic_params"])
my_baseline_params = baseline_parameters(
    basic_params=my_basic_params, **cfg["baseline_params"]
)
my_ad_spend_params = ad_spend_parameters(**cfg['ad_spend_params'])
my_media_params = media_parameters(**cfg["media_params"])
my_cvr_params = cvr_parameters(**cfg["cvr_params"])
my_adstock_params = adstock_parameters(**cfg["adstock_params"])
my_output_params = output_parameters(**cfg["output_params"])



from pysimmmulator import load_parameters

def test_load_cfg():
    assert type(load_parameters.cfg) == dict

def test_cfg_base_keys():
    cfg = load_parameters.cfg
    assert 'basic_params' in cfg.keys()

def test_media_cfg_check():
    cfg = load_parameters.cfg
    media_params = load_parameters.media_parameters(**cfg['media_params'])
    media_params.check(basic_params=load_parameters.my_basic_params)
    assert True == True

def test_cvr_cfg_check():
    cfg = load_parameters.cfg
    cvr_params = load_parameters.cvr_parameters(**cfg['cvr_params'])
    cvr_params.check(basic_params=load_parameters.my_basic_params)
    assert True == True

def test_adstock_cfg_check():
    cfg = load_parameters.cfg
    adstock_params = load_parameters.adstock_parameters(**cfg['adstock_params'])
    adstock_params.check(basic_params=load_parameters.my_basic_params)
    assert True == True


from dataclasses import dataclass
import datetime


@dataclass
class basic_parameters:
    """Handler for loading in basic parameters used by simmmulate class.
    After init, this class will also preform logic checks for the values of the 
    parameters.
     
    Args:
        years (int): Number of years you want to generate data for.
        channels_impressions (list[int]): names of media channels that use impressions as their metric of activity (Examples: Amazon, TV, etc)
        channels_clicks (list[int]): names of media channels that use clicks as their metric of activity (Examples: Search)
        frequency_of_campaigns (int): how often campaigns occur (for example, frequency of 1 would yield a new campaign every 1 day with each campaign lasting 1 day).
        start_date (str): format yyyy/mm/dd that determines when your daily data set starts on
        true_cvr (list): what the underlying conversion rates of all the channels are, statistical noise will be added on top of this.
        revenue_per_conv (float): How much money we make from a conversion (i.e. profit from a unit of sale)."""
    years: int
    channels_impressions: list[str]
    channels_clicks: list[str]
    frequency_of_campaigns: int
    start_date: str
    true_cvr: list = None
    revenue_per_conv: float = None

    def __post_init__(self):
        self.all_channels = self.channels_clicks + self.channels_impressions
        self.start_date = datetime.datetime.strptime(self.start_date, "%Y/%m/%d")
        self.end_date = self.start_date + datetime.timedelta(days=(self.years * 365))
        self.evaluate_params()

    def evaluate_params(self):
        assert (
            self.years > 0
        ), "You entered less than 1 year. Must generate more than a years worth of data"
        if self.true_cvr is not None:
            assert len(self.true_cvr.keys()) == len(
                self.all_channels
            ), "True CVR must have equal number of entries as channel impressions and channel clicks"
            for cvr in self.true_cvr.values():
                assert (
                    0 < cvr <= 1
                ), "You've entered an invalid True CVR value. CVR values must be greater than 0 and less than or equal to 1"
        assert (
            self.frequency_of_campaigns >= 1
        ), "You entered a frequency of campaigns less than 1. You must enter a number greater than 1"

    def __repr__(self):
        channel_use_impressions = ", ".join(self.channels_impressions)
        channel_use_clicks = ", ".join(self.channels_clicks)
        cvr_values = (
            ", ".join([str(cvr) for cvr in self.true_cvr])
            if self.true_cvr is not None
            else ""
        )

        slug = f"""Years of Data to generate : {self.years}
Channel that use impressions : {channel_use_impressions}
Channel that use clicks : {channel_use_clicks}
How frequently campaigns occur : {self.frequency_of_campaigns}
True CVRs of a channel (in order of channels you specified) : {cvr_values}
Revenue per conversion : {self.revenue_per_conv}
Date the data set will start with : {self.start_date}"""

        return slug


@dataclass
class baseline_parameters:
    """Handler for loading in parameters used by simmmulate class to generate a baseline of sales.
    After init, this class will also preform logic checks for the values of the 
    parameters.
     
    Args:
        basic_params (basic_parameters): Number of years you want to generate data for.
        base_p (int): Amount of baseline sales we get in a day (sales not due to ads)
        trend_p (int): How much baseline sales is going to grow over the whole period of our data.
        temp_var (int): How big the height of the sine function is for temperature -- i.e. how much temperature varies (used to inject seasonality into our data)
        temp_coef_mean (int): The average of how important seasonality is in our data (the larger this number, the more important seasonality is for sales)
        temp_coef_sd (int): The standard deviation of how important seasonality is in our data (the larger this number, the more variable the importance of seasonality is for sales)
        error_std (int): Amount of statistical noise added to baseline sales (the larger this number, the noisier baseline sales will be)."""
    basic_params: basic_parameters
    base_p: int
    trend_p: int
    temp_var: int
    temp_coef_mean: int
    temp_coef_sd: int
    error_std: int

    def __post_init__(self):
        assert (
            self.error_std < self.base_p
        ), "Error std can not exceed base sales value"


@dataclass
class ad_spend_parameters:
    """Handler for loading in parameters used by simmmulate class to generate ad spend approximation.
    After init, this class will also preform logic checks for the values of the 
    parameters. Also provided is a check function that when passed basic_params 
    from input to simmmulate, will provide futher validation checks.
     
    Args:
        campaign_spend_mean (int): The average amount of money spent on a campaign.
        campaign_spend_std (int): The standard deviation of money spent on a campaign
        max_min_proportion_on_each_channel (dict): Specifies the minimum and maximum percentages of total spend allocated to each channel."""
    campaign_spend_mean: int
    campaign_spend_std: int
    max_min_proportion_on_each_channel: dict

    def __post_init__(self):
        assert (self.campaign_spend_mean > 0), "You entered a negative average campaign spend. Enter a positive number."
        assert (
            self.campaign_spend_std < self.campaign_spend_mean
        ), "You've entered a campaign spend standard deviation larger than the mean."
        for k, v in self.max_min_proportion_on_each_channel.items():
            assert (
                0 < v["min"] <= 1
            ), "Min spend must be between 0 and 1 for each channel"
            assert (
                0 < v["max"] <= 1
            ), "Max spend must be between 0 and 1 for each channel"
    
    def check(self, basic_params: basic_parameters):
        """Validates ad_spend parameters agianst previously constructed basic 
        parameter values.
        
        Args:
            basic_params (basic_parameters): Previously submitted parameters as required by the simmmulate class"""
        assert (
            len(self.max_min_proportion_on_each_channel.keys()) - 1
            == len(basic_params.all_channels)
        ), "You did not input in enough numbers or put in too many numbers for proportion of spends on each channel. Must have a maximum and minimum percentage specified for all channels except the last channel, which will be auto calculated as any remaining amount."


@dataclass
class media_parameters:
    """Handler for loading in parameters used by simmmulate class to generate media data.
    After init, this class will also preform logic checks for the values of the 
    parameters. Also provided is a check function that when passed basic_params 
    from input to simmmulate, will provide futher validation checks.
     
    Args:
        true_cpm (dict): Specifies the true Cost per Impression (CPM) of each channel (noise will be added to this to simulate number of impressions)
        true_cpc (dict): Specifies the true Cost per Click (CPC) of each channel (noise will be added to this to simulate number of clicks)
        noisy_cpm_cpc (dict): Specifies the bias and scale of noise added to the true value CPM or CPC for each channel."""
    true_cpm: dict
    true_cpc: dict
    noisy_cpm_cpc: dict

    def __post_init__(self):
        self.true_cpmcpc_channels = list(self.true_cpm.keys()) + list(
            self.true_cpc.keys()
        )
        self.noise_channels = list(self.noisy_cpm_cpc.keys())

    def check(self, basic_params: basic_parameters):
        assert sorted(self.true_cpmcpc_channels) == sorted(
            basic_params.all_channels
        ), "Channels declared within true_cpm & true_cpc must be the same as original base channel input"
        for val in self.true_cpm.values():
            assert isinstance(val, float), "cpm values must be of type float"
            assert val > 0, "CPM values must be greater than 0"
        for val in self.true_cpc.values():
            assert isinstance(val, float), "cpc values must be of type float"
            assert val > 0, "CPC values must be greater than 0"

        assert sorted(self.noise_channels) == sorted(
            basic_params.all_channels
        ), "Channels declared within noisy_cpm_cpc must be the same as original base channel input"


@dataclass
class cvr_parameters:
    """Handler for loading in parameters used by simmmulate class to generate cvr data.
    Provided is a check function that when passed basic_params 
    from input to simmmulate, will provide validation checks.
     
    Args:
        noisy_cpm_cpc (dict): Specifies the bias and scale of noise added to the true value CVR for each channel."""
    noisy_cvr: dict

    def __post_init__(self):
        self.noise_channels = list(self.noisy_cvr.keys())

        for channel in self.noisy_cvr.keys():
            channel_spec = self.noisy_cvr[channel]
            assert isinstance(channel_spec['loc'], float), "noisy loc value must be of type float"
            assert isinstance(channel_spec['scale'], float), "noisy scale value must be of type float"

    def check(self, basic_params: basic_parameters):
        assert sorted(self.noise_channels) == sorted(
            basic_params.all_channels
        ), "Channels declared within noisy_cpm_cpc must be the same as original base channel input"


@dataclass
class adstock_parameters:
    """Handler for loading in parameters used by simmmulate class to augment adstock data.
    Provided is a check function that when passed basic_params 
    from input to simmmulate, will provide validation checks.
     
    Args:
        true_lambda_decay (dict): Numbers between 0 and 1 specifying the lambda parameters for a geometric distribution for adstocking media variables.
        alpha_saturation (dict): Specifying alpha parameter of geometric distribution for applying diminishing returns to media variables
        gamma_saturation (dict): Between 0 and 1 specifying gamma parameter of geometric distribution for applying diminishing returns to media variables
        """
    true_lambda_decay: dict
    alpha_saturation: dict
    gamma_saturation: dict

    def __post_init__(self):
        for channel, value in self.true_lambda_decay.items():
            assert isinstance(value, float), "lambda decay value must be of type float"
            assert 0 <= value <= 1, "lambda decay value must be between 0 and 1"
        for channel, value in self.alpha_saturation.items():
            assert isinstance(value, float), "alpha saturation value must be of type float"
        for channel, value in self.gamma_saturation.items():
            assert isinstance(value, float), "gamma saturation value must be of type float"
            assert 0 <= value <= 1, "gamma saturation value must be between 0 and 1"

    def check(self, basic_params: basic_parameters):
        for input_dict in [self.true_lambda_decay, self.alpha_saturation, self.gamma_saturation]:
            assert sorted(list(input_dict.keys())) == sorted(
                basic_params.all_channels
            ), f"Channels declared within {input_dict.__name__} must be the same as original base channel input"


@dataclass
class output_parameters:
    """Handler for loading in parameters used by simmmulate class to generate final output data.
     
    Args:
        aggregation_level (str): Specifies the aggregation level of final output data. choose between [daily, weekly]."""
    aggregation_level: str

    def __post_init__(self):
        assert self.aggregation_level in ['daily','weekly'], "{self.aggregation_level} is invalid. Aggregation level must be in [daily, weekly]"



from pysimmmulator import simmmulate, load_parameters

def test_initiate_sim():
    simmmulate.simulate(load_parameters.my_basic_params)

def test_step1_baseline():
    sim = simmmulate.simulate(load_parameters.my_basic_params)
    sim.simulate_baseline(**load_parameters.cfg["baseline_params"])

def test_step2_adspend():
    sim = simmmulate.simulate(load_parameters.my_basic_params)
    sim.simulate_baseline(**load_parameters.cfg["baseline_params"])
    sim.simulate_ad_spend(**load_parameters.cfg["ad_spend_params"])

def test_step3_media():
    sim = simmmulate.simulate(load_parameters.my_basic_params)
    sim.simulate_baseline(**load_parameters.cfg["baseline_params"])
    sim.simulate_ad_spend(**load_parameters.cfg["ad_spend_params"])
    sim.simulate_media(**load_parameters.cfg["media_params"])

def test_step4_cvr():
    sim = simmmulate.simulate(load_parameters.my_basic_params)
    sim.simulate_baseline(**load_parameters.cfg["baseline_params"])
    sim.simulate_ad_spend(**load_parameters.cfg["ad_spend_params"])
    sim.simulate_media(**load_parameters.cfg["media_params"])
    sim.simulate_cvr(**load_parameters.cfg["cvr_params"])

def tests_step5_adstock():
    sim = simmmulate.simulate(load_parameters.my_basic_params)
    sim.simulate_baseline(**load_parameters.cfg["baseline_params"])
    sim.simulate_ad_spend(**load_parameters.cfg["ad_spend_params"])
    sim.simulate_media(**load_parameters.cfg["media_params"])
    sim.simulate_cvr(**load_parameters.cfg["cvr_params"])
    sim.simulate_decay_returns(**load_parameters.cfg["adstock_params"])

def tests_step6_conversions():
    sim = simmmulate.simulate(load_parameters.my_basic_params)
    sim.simulate_baseline(**load_parameters.cfg["baseline_params"])
    sim.simulate_ad_spend(**load_parameters.cfg["ad_spend_params"])
    sim.simulate_media(**load_parameters.cfg["media_params"])
    sim.simulate_cvr(**load_parameters.cfg["cvr_params"])
    sim.simulate_decay_returns(**load_parameters.cfg["adstock_params"])
    sim.calculate_conversions()

def tests_step7_consolidatedataframe():
    sim = simmmulate.simulate(load_parameters.my_basic_params)
    sim.simulate_baseline(**load_parameters.cfg["baseline_params"])
    sim.simulate_ad_spend(**load_parameters.cfg["ad_spend_params"])
    sim.simulate_media(**load_parameters.cfg["media_params"])
    sim.simulate_cvr(**load_parameters.cfg["cvr_params"])
    sim.simulate_decay_returns(**load_parameters.cfg["adstock_params"])
    sim.calculate_conversions()
    sim.consolidate_dataframe()

def tests_step8_roi():
    sim = simmmulate.simulate(load_parameters.my_basic_params)
    sim.simulate_baseline(**load_parameters.cfg["baseline_params"])
    sim.simulate_ad_spend(**load_parameters.cfg["ad_spend_params"])
    sim.simulate_media(**load_parameters.cfg["media_params"])
    sim.simulate_cvr(**load_parameters.cfg["cvr_params"])
    sim.simulate_decay_returns(**load_parameters.cfg["adstock_params"])
    sim.calculate_conversions()
    sim.consolidate_dataframe()
    sim.calculate_channel_roi()

def tests_step9_consolidatedataframe():
    sim = simmmulate.simulate(load_parameters.my_basic_params)
    sim.simulate_baseline(**load_parameters.cfg["baseline_params"])
    sim.simulate_ad_spend(**load_parameters.cfg["ad_spend_params"])
    sim.simulate_media(**load_parameters.cfg["media_params"])
    sim.simulate_cvr(**load_parameters.cfg["cvr_params"])
    sim.simulate_decay_returns(**load_parameters.cfg["adstock_params"])
    sim.calculate_conversions()
    sim.consolidate_dataframe()
    sim.finalize_output(**load_parameters.cfg["output_params"])

def test_run_with_config():
    sim = simmmulate.simulate(load_parameters.my_basic_params)
    sim.run_with_config()


__title__ = 'CheetahPy'
__author__ = 'RyanAugust'
__license__ = 'MIT'
__copyright__ = 'Copyright 2024'

import os
with open(os.path.join('src/pysimmmulator', 'VERSION')) as version_file:
    __version__ = version_file.read().strip()